#jinja2: variable_start_string: '[%', variable_end_string: '%]'
apiVersion: unfurl/v1alpha1
kind: Project
[%include | default("") %]
{%if manifestPath %}
manifests:
  - file: [%manifestPath%]
    default: true
{% endif %}
