apiVersion: unfurl/v1alpha1
kind: Ensemble
spec:
  service_template:
    topology_template:
      node_templates:
        # nodeExample:
        #   type: tosca.nodes.Root
        #   properties:
        #     # you can add define adhoc properties here:
        #     # example: "example"
        #   interfaces:
        #     Standard:
        #       operations:
        #         create:
        #           implementation: unfurl.configurators.ansible.AnsibleConfigurator
        #           inputs:
        #             playbook:
        #               q: # quote so that jinja2 expressions aren't evaluated util later by ansible
        #                  # define an ansible playbook here, it will run a OPERATION_HOST (default: localhost)
